#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#
import time
from collections.abc import Mapping as ABCMapping
from typing import Any, Mapping, Optional

from airbyte_cdk.models import AirbyteLogMessage, AirbyteMessage, AirbyteRecordMessage, AirbyteTraceMessage
from airbyte_cdk.models import Type as MessageType
from airbyte_cdk.models.file_transfer_record_message import AirbyteFileTransferRecordMessage
from airbyte_cdk.sources.streams.core import StreamData
from airbyte_cdk.sources.utils.transform import TransformConfig, TypeTransformer


def stream_data_to_airbyte_message(
    stream_name: str,
    data_or_message: StreamData,
    transformer: TypeTransformer = TypeTransformer(TransformConfig.NoTransform),
    schema: Optional[Mapping[str, Any]] = None,
    is_file_transfer_message: bool = False,
) -> AirbyteMessage:
    if schema is None:
        schema = {}

    match data_or_message:
        case ABCMapping():
            data = dict(data_or_message)
            now_millis = time.time_ns() // 1_000_000
            # Transform object fields according to config. Most likely you will
            # need it to normalize values against json schema. By default no action
            # taken unless configured. See
            # docs/connector-development/cdk-python/schemas.md for details.
            transformer.transform(data, schema)  # type: ignore
            if is_file_transfer_message:
                message = AirbyteFileTransferRecordMessage(stream=stream_name, file=data, emitted_at=now_millis, data={})
            else:
                message = AirbyteRecordMessage(stream=stream_name, data=data, emitted_at=now_millis)
            return AirbyteMessage(type=MessageType.RECORD, record=message)
        case AirbyteTraceMessage():
            return AirbyteMessage(type=MessageType.TRACE, trace=data_or_message)
        case AirbyteLogMessage():
            return AirbyteMessage(type=MessageType.LOG, log=data_or_message)
        case _:
            raise ValueError(f"Unexpected type for data_or_message: {type(data_or_message)}: {data_or_message}")



# Copyright (c) 2024 Airbyte, Inc., all rights reserved.
"""A MotherDuck implementation of the cache, built on DuckDB.

## Usage Example

```python
from airbyte as ab
from airbyte.caches import MotherDuckCache

cache = MotherDuckCache(
    database="mydatabase",
    schema_name="myschema",
    api_key=ab.get_secret("MOTHERDUCK_API_KEY"),
)
"""

from __future__ import annotations

import warnings

from airbyte_cdk.sql._processors.duckdb import DuckDBConfig, DuckDBSqlProcessor
from airbyte_cdk.sql.secrets import SecretString
from duckdb_engine import DuckDBEngineWarning
from overrides import overrides
from pydantic import Field

# Suppress warnings from DuckDB about reflection on indices.
# https://github.com/Mause/duckdb_engine/issues/905
warnings.filterwarnings(
    "ignore",
    message="duckdb-engine doesn't yet support reflection on indices",
    category=DuckDBEngineWarning,
)


class MotherDuckConfig(DuckDBConfig):
    """Configuration for the MotherDuck cache."""

    database: str = Field()
    api_key: SecretString = Field()
    db_path: str = Field(default="md:")
    custom_user_agent: str = Field(default="airbyte")

    @overrides
    def get_sql_alchemy_url(self) -> SecretString:
        """Return the SQLAlchemy URL to use."""
        # Suppress warnings from DuckDB about reflection on indices.
        # https://github.com/Mause/duckdb_engine/issues/905
        warnings.filterwarnings(
            "ignore",
            message="duckdb-engine doesn't yet support reflection on indices",
            category=DuckDBEngineWarning,
        )

        return SecretString(
            f"duckdb:///md:{self.database}?motherduck_token={self.api_key}"
            f"&custom_user_agent={self.custom_user_agent}"
            # Not sure why this doesn't work. We have to override later in the flow.
            # f"&schema={self.schema_name}"
        )

    @overrides
    def get_database_name(self) -> str:
        """Return the name of the database."""
        return self.database


class MotherDuckSqlProcessor(DuckDBSqlProcessor):
    """A cache implementation for MotherDuck."""

    supports_merge_insert = False

    @overrides
    def _setup(self) -> None:
        """Do any necessary setup, if applicable.

        Note: The DuckDB parent class requires pre-creation of local directory structure. We
        don't need to do that here so we override the method be a no-op.
        """
        # No setup to do and no need to pre-create local file storage.
        pass



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#
import logging
from io import IOBase
from typing import Any, Dict, Generator, Iterable, Optional, Tuple

from airbyte_cdk.sources.file_based.config.file_based_stream_config import FileBasedStreamConfig
from airbyte_cdk.sources.file_based.file_based_stream_reader import AbstractFileBasedStreamReader, FileReadMode
from airbyte_cdk.sources.file_based.remote_file import RemoteFile
from airbyte_cdk.sources.file_based.writers.local_file_client import LocalFileTransferClient


class _FileReader:
    def read_data(
        self,
        config: FileBasedStreamConfig,
        file: RemoteFile,
        stream_reader: AbstractFileBasedStreamReader,
        logger: logging.Logger,
        file_read_mode: FileReadMode,
    ) -> Generator[Tuple[IOBase, int], None, None]:

        try:
            file_size = stream_reader.file_size(file)
            with stream_reader.open_file(file, file_read_mode, "UTF-8", logger) as fp:
                yield fp, file_size

        except Exception as ex:
            logger.error("An error has occurred while reading file: %s", str(ex))


class BlobTransfer:
    def __init__(self, file_reader: Optional[_FileReader] = None):
        self._file_reader = file_reader if file_reader else _FileReader()

    def write_streams(
        self,
        config: FileBasedStreamConfig,
        file: RemoteFile,
        stream_reader: AbstractFileBasedStreamReader,
        logger: logging.Logger,
    ) -> Iterable[Dict[str, Any]]:
        file_no = 0
        try:
            data_generator = self._file_reader.read_data(config, file, stream_reader, logger, self.file_read_mode)
            local_writer = LocalFileTransferClient()
            for file_opened, file_size in data_generator:
                yield local_writer.write(file.uri, file_opened, file_size, logger)
                file_no += 1
        except Exception as ex:
            logger.error("An error has occurred while writing file: %s", str(ex))
            raise ex
        finally:
            data_generator.close()

    @property
    def file_read_mode(self) -> FileReadMode:
        return FileReadMode.READ



# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
"""A DuckDB implementation of the cache."""

from __future__ import annotations

import logging
import warnings
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, List, Literal

import pyarrow as pa
from airbyte_cdk import DestinationSyncMode
from airbyte_cdk.sql import exceptions as exc
from airbyte_cdk.sql.constants import AB_EXTRACTED_AT_COLUMN
from airbyte_cdk.sql.secrets import SecretString
from airbyte_cdk.sql.shared.sql_processor import SqlConfig, SqlProcessorBase, SQLRuntimeError
from duckdb_engine import DuckDBEngineWarning
from overrides import overrides
from pydantic import Field
from sqlalchemy import Executable, TextClause, text
from sqlalchemy.exc import ProgrammingError, SQLAlchemyError

if TYPE_CHECKING:
    from sqlalchemy.engine import Connection, Engine

logger = logging.getLogger(__name__)


# @dataclass
class DuckDBConfig(SqlConfig):
    """Configuration for DuckDB."""

    db_path: Path | str = Field()
    """Normally db_path is a Path object.

    The database name will be inferred from the file name. For example, given a `db_path` of
    `/path/to/my/duckdb-file`, the database name is `my_db`.
    """

    schema_name: str = Field(default="main")
    """The name of the schema to write to. Defaults to "main"."""

    @overrides
    def get_sql_alchemy_url(self) -> SecretString:
        """Return the SQLAlchemy URL to use."""
        # Suppress warnings from DuckDB about reflection on indices.
        # https://github.com/Mause/duckdb_engine/issues/905
        warnings.filterwarnings(
            "ignore",
            message="duckdb-engine doesn't yet support reflection on indices",
            category=DuckDBEngineWarning,
        )
        return SecretString(f"duckdb:///{self.db_path!s}")

    @overrides
    def get_database_name(self) -> str:
        """Return the name of the database."""
        if self.db_path == ":memory:":
            return "memory"

        # Split the path on the appropriate separator ("/" or "\")
        split_on: Literal["/", "\\"] = "\\" if "\\" in str(self.db_path) else "/"

        # Return the file name without the extension
        return str(self.db_path).split(sep=split_on)[-1].split(".")[0]

    def _is_file_based_db(self) -> bool:
        """Return whether the database is file-based."""
        if isinstance(self.db_path, Path):
            return True

        db_path_str = str(self.db_path)
        return (
            ("/" in db_path_str or "\\" in db_path_str)
            and db_path_str != ":memory:"
            and "md:" not in db_path_str
            and "motherduck:" not in db_path_str
        )

    @overrides
    def get_sql_engine(self) -> Engine:
        """Return the SQL Alchemy engine.

        This method is overridden to ensure that the database parent directory is created if it
        doesn't exist.
        """
        if self._is_file_based_db():
            Path(self.db_path).parent.mkdir(parents=True, exist_ok=True)

        return super().get_sql_engine()


class DuckDBSqlProcessor(SqlProcessorBase):
    """A DuckDB implementation of the cache.

    Jsonl is used for local file storage before bulk loading.
    Unlike the Snowflake implementation, we can't use the COPY command to load data
    so we insert as values instead.
    """

    supports_merge_insert = False
    sql_config: DuckDBConfig

    @overrides
    def _setup(self) -> None:
        """Create the database parent folder if it doesn't yet exist."""
        if self.sql_config.db_path == ":memory:":
            return

        Path(self.sql_config.db_path).parent.mkdir(parents=True, exist_ok=True)

    def _create_table_if_not_exists(
        self,
        table_name: str,
        column_definition_str: str,
        primary_keys: list[str] | None = None,
    ) -> None:
        if primary_keys:
            pk_str = ", ".join(primary_keys)
            column_definition_str += f",\n  PRIMARY KEY ({pk_str})"

        cmd = f"""
        CREATE TABLE IF NOT EXISTS {self._fully_qualified(table_name)} (
            {column_definition_str}
        )
        """
        _ = self._execute_sql(cmd)

    def _do_checkpoint(
        self,
        connection: Connection | None = None,
    ) -> None:
        """Checkpoint the given connection.

        We override this method to ensure that the DuckDB WAL is checkpointed explicitly.
        Otherwise DuckDB will lazily flush the WAL to disk, which can cause issues for users
        who want to manipulate the DB files after writing them.

        For more info:
        - https://duckdb.org/docs/sql/statements/checkpoint.html
        """
        if connection is not None:
            connection.execute(text("CHECKPOINT"))
            return

        with self.get_sql_connection() as new_conn:
            new_conn.execute(text("CHECKPOINT"))

    def _executemany(self, sql: str | TextClause | Executable, params: list[list[Any]]) -> None:
        """Execute the given SQL statement."""
        if isinstance(sql, str):
            sql = text(sql)

        with self.get_sql_connection() as conn:
            try:
                entries = list(params)
                conn.engine.pool.connect().executemany(str(sql), entries)  # type: ignore
            except (
                ProgrammingError,
                SQLAlchemyError,
            ) as ex:
                msg = f"Error when executing SQL:\n{sql}\n{type(ex).__name__}{ex!s}"
                raise SQLRuntimeError(msg) from None  # from ex

    def _write_with_executemany(self, buffer: Dict[str, Dict[str, List[Any]]], stream_name: str, table_name: str) -> None:
        column_names_list = list(buffer[stream_name].keys())
        column_names = ", ".join(column_names_list)
        params = ", ".join(["?"] * len(column_names_list))
        sql = f"""
        -- Write with executemany
        INSERT INTO {self._fully_qualified(table_name)}
            ({column_names})
        VALUES ({params})
        """
        entries_to_write = buffer[stream_name]
        num_entries = len(entries_to_write[column_names_list[0]])
        parameters = [[entries_to_write[column_name][n] for column_name in column_names_list] for n in range(num_entries)]
        self._executemany(sql, parameters)

    def _write_from_pa_table(self, table_name: str, stream_name: str, pa_table: pa.Table) -> None:
        full_table_name = self._fully_qualified(table_name)
        columns = list(self._get_sql_column_definitions(stream_name).keys())
        if len(columns) != len(pa_table.column_names):
            warnings.warn(f"Schema has colums: {columns}, buffer has columns: {pa_table.column_names}")
        column_names = ", ".join(pa_table.column_names)
        sql = f"""
        -- Write from PyArrow table
        INSERT INTO {full_table_name} ({column_names}) SELECT {column_names} FROM pa_table
        """
        self._execute_sql(sql)

    def _write_temp_table_to_target_table(
        self,
        stream_name: str,
        temp_table_name: str,
        final_table_name: str,
        sync_mode: DestinationSyncMode,
    ) -> None:
        """Write the temp table into the final table using the provided write strategy."""
        if sync_mode == DestinationSyncMode.overwrite:
            # Note: No need to check for schema compatibility
            # here, because we are fully replacing the table.
            self._swap_temp_table_with_final_table(
                stream_name=stream_name,
                temp_table_name=temp_table_name,
                final_table_name=final_table_name,
            )
            return

        if sync_mode == DestinationSyncMode.append:
            self._ensure_compatible_table_schema(
                stream_name=stream_name,
                table_name=final_table_name,
            )
            self._append_temp_table_to_final_table(
                stream_name=stream_name,
                temp_table_name=temp_table_name,
                final_table_name=final_table_name,
            )
            return

        if sync_mode == DestinationSyncMode.append_dedup:
            self._ensure_compatible_table_schema(
                stream_name=stream_name,
                table_name=final_table_name,
            )
            if not self.supports_merge_insert:
                # Fallback to emulated merge if the database does not support merge natively.
                self._emulated_merge_temp_table_to_final_table(
                    stream_name=stream_name,
                    temp_table_name=temp_table_name,
                    final_table_name=final_table_name,
                )
                return

            self._merge_temp_table_to_final_table(
                stream_name=stream_name,
                temp_table_name=temp_table_name,
                final_table_name=final_table_name,
            )
            return

        raise exc.AirbyteInternalError(
            message="Sync mode is not supported.",
            context={
                "sync_mode": sync_mode,
            },
        )

    def _drop_duplicates(self, table_name: str, stream_name: str) -> str:
        primary_keys = self.catalog_provider.get_primary_keys(stream_name)
        new_table_name = f"{table_name}_deduped"
        if primary_keys:
            pks = ", ".join(primary_keys)
            sql = f"""
            -- Drop duplicates from temp table
            CREATE TABLE {self._fully_qualified(new_table_name)} AS (
                SELECT * FROM {self._fully_qualified(table_name)}
                QUALIFY row_number() OVER (PARTITION BY ({pks}) ORDER BY {AB_EXTRACTED_AT_COLUMN} DESC) = 1
            )
            """
            self._execute_sql(sql)
            return new_table_name
        return table_name

    def write_stream_data_from_buffer(
        self,
        buffer: Dict[str, Dict[str, List[Any]]],
        stream_name: str,
        sync_mode: DestinationSyncMode,
    ) -> None:
        temp_table_name = self._create_table_for_loading(stream_name, batch_id=None)
        try:
            pa_table = pa.Table.from_pydict(buffer[stream_name])
        except Exception:
            logger.exception(
                "Writing with PyArrow table failed, falling back to writing with executemany. Expect some performance degradation."
            )
            self._write_with_executemany(buffer, stream_name, temp_table_name)
        else:
            # DuckDB will automatically find and SELECT from the `pa_table`
            # local variable defined above.
            self._write_from_pa_table(temp_table_name, stream_name, pa_table)

        temp_table_name_dedup = self._drop_duplicates(temp_table_name, stream_name)

        try:
            self._write_temp_table_to_target_table(
                stream_name=stream_name,
                temp_table_name=temp_table_name_dedup,
                final_table_name=stream_name,
                sync_mode=sync_mode,
            )
        finally:
            self._drop_temp_table(temp_table_name_dedup, if_exists=True)
            self._drop_temp_table(temp_table_name, if_exists=True)



# Copyright (c) 2024 Airbyte, Inc., all rights reserved.

import logging
import os
import time
from io import IOBase
from typing import Any, Dict

import psutil
from airbyte_cdk.models import FailureType
from airbyte_cdk.sources.file_based.exceptions import FileSizeLimitError

AIRBYTE_STAGING_DIRECTORY = os.getenv("AIRBYTE_STAGING_DIRECTORY", "/staging/files")
DEFAULT_LOCAL_DIRECTORY = "/tmp/airbyte-file-transfer"


class LocalFileTransferClient:
    FILE_SIZE_LIMIT = 1_000_000_000

    def __init__(self) -> None:
        """
        Initialize the LocalFileTransferClient. It uses a default local directory for file saving.
        """
        self._local_directory = AIRBYTE_STAGING_DIRECTORY if os.path.exists(AIRBYTE_STAGING_DIRECTORY) else DEFAULT_LOCAL_DIRECTORY

    def write(self, file_uri: str, fp: IOBase, file_size: int, logger: logging.Logger) -> Dict[str, Any]:
        """
        Write the file to a local directory.
        """
        if file_size > self.FILE_SIZE_LIMIT:
            message = "File size exceeds the 1 GB limit."
            raise FileSizeLimitError(message=message, internal_message=message, failure_type=FailureType.config_error)

        # Remove left slashes from source path format to make relative path for writing locally
        file_relative_path = file_uri.lstrip("/")
        local_file_path = os.path.join(self._local_directory, file_relative_path)

        # Ensure the local directory exists
        os.makedirs(os.path.dirname(local_file_path), exist_ok=True)

        # Get the absolute path
        absolute_file_path = os.path.abspath(local_file_path)

        # Get available disk space
        disk_usage = psutil.disk_usage("/")
        available_disk_space = disk_usage.free

        # Get available memory
        memory_info = psutil.virtual_memory()
        available_memory = memory_info.available

        # logger.info(f"Writing file to {local_file_path}.")
        # Log file size, available disk space, and memory
        logger.info(
            f"Writing file to '{local_file_path}' "
            f"with size: {file_size / (1024 * 1024):,.2f} MB ({file_size / (1024 * 1024 * 1024):.2f} GB), "
            f"available disk space: {available_disk_space / (1024 * 1024):,.2f} MB ({available_disk_space / (1024 * 1024 * 1024):.2f} GB),"
            f"available memory: {available_memory / (1024 * 1024):,.2f} MB ({available_memory / (1024 * 1024 * 1024):.2f} GB)."
        )

        with open(local_file_path, "wb") as f:
            # Measure the time for reading
            logger.info("Starting to read the file")
            start_read_time = time.time()
            # todo: read chunks or allow reader to implement their own download process https://github.com/airbytehq/airbyte-internal-issues/issues/10480
            file_content = fp.read()  # Read the file content
            read_duration = time.time() - start_read_time
            logger.info(f"Time taken to read the file: {read_duration:,.2f} seconds.")

            # Measure the time for writing
            logger.info("Starting to write the file locally")
            start_write_time = time.time()
            f.write(file_content)
            write_duration = time.time() - start_write_time
            logger.info(f"Time taken to write the file: {write_duration:,.2f} seconds.")

        logger.info(f"File {file_relative_path} successfully written to {self._local_directory}.")

        return {"file_url": absolute_file_path, "bytes": file_size, "file_relative_path": file_relative_path}
